import re
import json


def parse(text):
    matcher_json = r'\{[\s\S]*?\}'  # 仕様に正確に従っているわけではないので注意
    json_strings = re.findall(matcher_json, text)

    json_data = []
    for json_string in json_strings:
        try:
            json_data += [json.loads(json_string)]
        except json.JSONDecodeError:
            print(f'decode error: {json_string}')
    return json_data



# %%
import os
from anthropic import Anthropic


class AnthropicLLM:
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        if "model" not in kwargs:
            self.kwargs["model"] = "claude-3-5-sonnet-20240620"
        if "max_tokens" not in kwargs:
            self.kwargs["max_tokens"] = 200
        if "stream" not in kwargs:
            self.kwargs["stream"] = False
        self.client = Anthropic(api_key=os.getenv("ANTHROPIC_API_KEY"))

    def __call__(self, input_text):
        messages = [{"role": "user", "content": input_text}]
        response = self.client.messages.create(messages=messages, **self.kwargs)
        output_text = response.content[0].text
        return output_text


if __name__ == "__main__":
    llm = AnthropicLLM()
    print(llm("Hello, world!"))



import re


def parse(text):
    matcher_tag = r'(<([^>\s\n]+)([^>\n]*)?>)'
    cursor = 0
    tree = []
    text_split = re.split(matcher_tag, text)
    while cursor < len(text_split):
        if re.match(matcher_tag, text_split[cursor]) is None:
            tree.append(text_split[cursor])
            cursor += 1
            continue

        raw, tag, attributes = text_split[cursor:cursor+3]
        attributes = dict(re.findall(r'(\S*?)="(.*?)"', attributes))

        if '/' in tag:
            for j in range(len(tree)-1, -1, -1):
                if not isinstance(tree[j], dict):
                    continue
                if tree[j]['tag'] != tag[1:]:
                    continue
                tree[j]['content'] = tree[j+1:]
                tree = tree[:j+1]
                cursor += 3
                break
            continue

        element = {'tag': tag, 'attributes': attributes, 'content': None}
        tree.append(element)
        cursor += 3
    return tree


def _repr_tag(tag, content=None, **kwargs):
    if kwargs:
        attr = ' ' + ' '.join([f'{k}="{v}"' for k, v in kwargs.items()])
    else:
        attr = ''
    bra = f'<{tag}{attr}>'
    ket = f'</{tag}>'
    if content is None:
        return bra
    return ''.join([bra, content, ket])


def deparse(tree):
    if tree is None:
        return tree

    text = ''

    for element in tree:
        if isinstance(element, str):
            text += element
            continue
        element['content'] = deparse(element['content'])
        text += _repr_tag(element['tag'], element['content'],
                          **element['attributes'])
    return text


def findall(tree, tag):
    if tree is None:
        return []

    result = []
    for element in tree:
        if not isinstance(element, dict):
            continue
        if element['tag'] == tag:
            result.append(element)
        result += findall(element['content'], tag)
    return result


def normalize(tree, tags, outermost=True):
    if tree is None:
        return tree

    normalized = []
    for element in tree:
        if isinstance(element, str):
            if not outermost:
                normalized.append(element)
            continue
        if element['tag'] not in tags:
            continue
        element['content'] = normalize(element['content'], tags, False)
        normalized.append(element)
        if outermost:
            normalized.append('\n\n')
    return normalized



from .parser import xml


class LLMLinkBase:
    """LLMLinkの基底クラス
    """

    def __init__(self, llm, prompt_template, *args, docstring=None, **kwargs):
        self.llm = llm
        self.prompt_template = prompt_template
        if docstring is not None:
            self.__doc__ = docstring

    def format(self, *args, **kwargs):
        raise NotImplementedError

    def parse(self, text):
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        prompt = self.format(*args, **kwargs)
        ret = self.llm(prompt)
        outputs = self.parse(ret)
        return outputs


class LLMLink(LLMLinkBase):
    # Todo: SimpleLLMLink などに改名する（互換性のため現状そのまま）
    def __init__(
        self,
        llm,
        prompt_template,
        input_variables,
        output_variables,
        docstring=None
    ):
        super().__init__(llm, prompt_template, docstring=docstring)
        self.input_variables = input_variables
        self.output_variables = output_variables

    def format(self, **kwargs):
        inputs = {var: "None" for var in self.input_variables}
        inputs.update(kwargs)
        prompt = self.prompt_template.format(**kwargs)
        return prompt

    def parse(self, text):
        outputs = {var: [] for var in self.output_variables}
        for var in self.output_variables:
            for leaf in xml.findall(xml.parse(text), var):
                outputs[var].append(xml.deparse(leaf["content"]).strip())
        return outputs



# %%
from llm_models.openai_model import OpenAILLM
from llm_models.google_model import GoogleLLM
from llm_models.anthropic_model import AnthropicLLM


class LLMClient:
    """
    様々な大規模言語モデルと対話するためのクライアントを初期化し、対話します。
    """

    def __init__(self, llm_name):
        self.llm = self.initialize_llm(llm_name)

    def initialize_llm(self, llm_name):
        """
        提供されたモデル名に基づいて大規模言語モデルを初期化します。
        """
        model_map = {
            "gpt-4o-2024-08-06": OpenAILLM,
            "gpt-4o-2024-05-13": OpenAILLM,
            "gpt-4o-mini-2024-07-18": OpenAILLM,
            "gpt-4-turbo-2024-04-09": OpenAILLM,
            "gpt-4-0125-preview": OpenAILLM,
            "o1-preview-2024-09-12": OpenAILLM,
            "o1-mini-2024-09-12": OpenAILLM,
            "gemini-1.0-pro": GoogleLLM,
            "gemini-1.5-pro": GoogleLLM,
            "gemini-1.5-flash": GoogleLLM,
            "claude-3-5-sonnet-20240620": AnthropicLLM,
            "claude-3-opus-20240229": AnthropicLLM,
        }
        if llm_name in model_map:
            return model_map[llm_name](model=llm_name)
        else:
            raise ValueError(f"未知のLLM: {llm_name}")

    def __call__(self, input_text):
        """
        クライアントインスタンスが呼び出されたときに実行される機能を実装します。
        """
        return self.llm(input_text)


if __name__ == "__main__":
    llm_name = "gpt-4o-2024-08-06"
    llm_name = "o1-preview-2024-09-12"
    llm = LLMClient(llm_name)
    print(llm("Hello, world!"))



import tomllib
import tomli_w
from .parser import xml
from .link import LLMLinkBase


PROMPT_COMPILER = '''
<RULE>
All texts must be formatted in XML format. XML element ::= <tag attribute="value">content</tag>
Tags determine the meaning and function of the content. The content must not contradict the definition of the tag.
</RULE>

<TAG name="RULE">
This tag defines rules. The defined content is absolute.
</TAG>

<TAG name="TAG">
This tag defines a tag. The defined content is absolute.
Attributes:
    - name : A tag name.
</TAG>

<TAG name="THINK">
This tag is used to describe the thought process.
The thought process must be described in detail step-by-step.
Attributes:
    - label (optional) : A label summarizing the contents.
</TAG>

<TAG name="LIST">
This tag is used to describe a collection of the same XML elements.
Multiple identical tags are included within this tag.
Two or more types of tags must not exist within the LIST tag.
<EXAMPLE>
<LIST>
<ELEMENT>content1</ELEMENT>
<ELEMENT>content2</ELEMENT>
<ELEMENT>content3</ELEMENT>
</LIST>
</EXAMPLE>
</TAG>

<TAG name="EXAMPLE">
This tag describes an example.
The content described within this tag does not affect anything outside this tag.
For example, tags defined within the EXAMPLE tag are invalid outside the tag.
</TAG>

<TAG name="PROCESS">
This tag describes the information processing process.
The assistant must perform information processing according to the content of this tag.
</PROCESS>

<TAG name="INPUT_TAGS">
Describes the correspondence between input (arguments) and tags.
The correspondence is described as
argument1=TAG1,argument2=TAG2,...
separated by commas.
</TAG>

<TAG name="OUTPUT_TAGS">
Describes the correspondence between output (returns) and tags.
The correspondence is described as
return1=TAG1,return2=TAG2,...
separated by commas.
Notes:
    - For array-type returns (those in the form List[element]), the return should correspond to the element tag of the LIST tag, not the LIST tag itself.
</TAG>

<TAG name="SOURCE">
This tag describes the specifications of the information processing process.
The information processing process is described in the form of a python docstring.
The information processing process consists of input (arguments), output (returns), and processing content.
</TAG>

<RULE role="assistant">
The assistant executes the following process for the SOURCE.
<PROCESS>
1. The assistant carefully considers the content of the given SOURCE using the THINK tag. The assistant contemplates what steps are necessary to achieve the given process.
2. The assistant defines tags corresponding to the input (arguments) and output (returns) described in the SOURCE. The tag names are generally set to the same names as the arguments and returns described in the SOURCE, but different names can be used to avoid duplication.
3. If deemed effective, the assistant may define additional tags to assist in information processing.
4. The assistant uses INPUT_TAGS and OUTPUT_TAGS to map the defined tags to the arguments and returns in the SOURCE. Ensure accurate mapping.
5. The assistant describes the flow of information processing using the defined tags within the PROCESS tag.
</PROCESS>

Below is an example.
<EXAMPLE>
Suppose the following SOURCE is given to the assistant.
<SOURCE>
"""Split the text into words and count the number of words.

Args:
    text (str): Any text.

Returns:
    words (List[word]): Words contained in the input text.
    num_words (int): The number of words contained in the input text.

"""
</SOURCE>

In this case, the assistant converts the SOURCE into a combination of INPUT_TAGS, OUTPUT_TAGS, and PROCESS as follows.
<THINK label="Analysis of SOURCE">
The given SOURCE describes a process to split the input text into words and count the number of words.
First, a tag corresponding to the input text, text, needs to be defined.
Since the output words is a list of words, a word tag corresponding to a word should be defined and output within a LIST tag.
The granularity of "words" is not specified in the SOURCE. Therefore, here, words are defined as a collection of characters separated by spaces.
When counting words, to accurately determine the number,
</THINK>

<TAG name="TEXT">
This tag describes the text to be split into words.
</TAG>

<TAG name="WORD">
This tag describes a word.
Attributes:
    - count (optional) : Describes the order of the word.
</TAG>

<TAG name="NUM_WORDS">
This tag describes the number of words. Only numbers can be described within this tag.
</TAG>

<PROCESS>
1. Split the given string by spaces to form words.
2. Output the processing result as
<LIST>
<WORD count="1">word1</WORD>
<WORD count="2">word2</WORD>
...
</LIST>
3. Finally, output the number of words using the NUM_WORDS tag.
</PROCESS>

<INPUT_TAGS>
text=text
</INPUT_TAGS>

<OUTPUT_TAGS>
words=word,num_words=num_words
</OUTPUT_TAGS>

</EXAMPLE>
</RULE>

<SOURCE>
{source}
</SOURCE>
'''.strip()
OUTPUT_VARIABLES = ['TAG', 'PROCESS', 'INPUT_TAGS', 'OUTPUT_TAGS']


PROMPT_BASE = '''
<RULE>
All texts must be formatted in XML format. XML element ::= <tag attribute="value">content</tag>
Tags determine the meaning and function of the content. The content must not contradict the definition of the tag.
</RULE>

<TAG name="RULE">
This tag defines rules. The defined content is absolute.
</TAG>

<TAG name="TAG">
This tag defines a tag. The defined content is absolute.
Attributes:
    - name : A tag name.
</TAG>

<TAG name="THINK">
This tag is used to describe the thought process.
The thought process must be described in detail step-by-step.
Attributes:
    - label (optional) : A label summarizing the contents.
</TAG>

<TAG name="LIST">
This tag is used to describe a collection of the same XML elements.
Multiple identical tags are included within this tag.
Two or more types of tags must not exist within the LIST tag.
<EXAMPLE>
<LIST>
<ELEMENT>content1</ELEMENT>
<ELEMENT>content2</ELEMENT>
<ELEMENT>content3</ELEMENT>
</LIST>
</EXAMPLE>
</TAG>

<TAG name="EXAMPLE">
This tag describes an example.
The content described within this tag does not affect anything outside this tag.
For example, tags defined within the EXAMPLE tag are invalid outside the tag.
</TAG>

<TAG name="PROCESS">
This tag describes the information processing process.
The assistant must perform information processing according to the content of this tag.
</PROCESS>

<RULE role="assistant">
The assistant must perform information processing according to the flow described in the PROCESS and output the necessary tags and their contents.
</RULE>
'''.strip()


class PromptGenerator(LLMLinkBase):

    def __init__(self, llm):
        super().__init__(llm, PROMPT_COMPILER)

    def format(self, source):
        prompt = self.prompt_template.format(source=source)
        return prompt

    def parse(self, text):
        output = {var: [] for var in OUTPUT_VARIABLES}
        for var in OUTPUT_VARIABLES:
            for leaf in xml.findall(xml.parse(text), var):
                output[var].append(xml.deparse([leaf]).strip())
        return output


class LLMCompiler:

    def __init__(self, llm):
        self.prompt_generator = PromptGenerator(llm)

    def compile(self, source):
        ret = self.prompt_generator(source)

        assert len(ret['INPUT_TAGS']) == 1
        assert len(ret['OUTPUT_TAGS']) == 1

        input_tags = xml.parse(ret['INPUT_TAGS'][0])[1]['content'][0].strip().split(',')
        input_tags = dict(map(lambda x:x.split('='), input_tags))
        output_tags = xml.parse(ret['OUTPUT_TAGS'][0])[1]['content'][0].strip().split(',')
        output_tags = dict(map(lambda x:x.split('='), output_tags))
    
        prompt = PROMPT_BASE
        prompt += '\n\n' + '\n\n'.join(ret['TAG'])
        prompt += '\n\n' + '\n\n'.join(ret['PROCESS'])

        # for arg, tag in input_tags.items():
        #     prompt += f'\n\n<{tag}>'
        #     prompt += '\n{' + arg + '}'
        #     prompt += f'\n<\\{tag}>'

        params = {
            'source': source,
            'prompt_template': prompt,
            'input_tags': input_tags,
            'output_tags': output_tags
        }

        compiled = tomli_w.dumps(params, multiline_strings=True)
        return compiled


class CompiledLLMLink(LLMLinkBase):

    def __init__(self, llm, compiled):
        params = tomllib.loads(compiled)
        super().__init__(llm, params['prompt_template'], docstring=params['source'])
        self.input_tags = params['input_tags']
        self.output_tags = params['output_tags']

    def format(self, **kwargs):
        prompt = self.prompt_template

        for arg in kwargs:
            if arg not in self.input_tags:
                continue

            tag = self.input_tags[arg]
            values = kwargs[arg]
            if not isinstance(values, list):
                values = [values]

            prompt += '\n'
            if len(values) > 1:
                prompt += '\n<LIST>'
            for value in values:
                prompt += f'\n<{tag}>'
                prompt += f'\n{value}'
                prompt += f'\n</{tag}>'
            if len(values) > 1:
                prompt += '\n</LIST>'
        return prompt

    def parse(self, text):
        outputs = {var: [] for var in self.output_tags}
        for var in self.output_tags:
            tag = self.output_tags[var]
            for leaf in xml.findall(xml.parse(text), tag):
                outputs[var].append(xml.deparse(leaf["content"]).strip())
        return outputs
